@@ -148,19 +148,7 @@ khm_process_query_app_ver(khm_query_app_version * papp_ver) {
     if (papp_ver->request_swap) {
         khm_version ver_caller = papp_ver->ver_caller;
 
-        if (ver_caller.major >   app_version.major ||
-
-            (ver_caller.major == app_version.major &&
-             ver_caller.minor >  app_version.minor) ||
-
-            (ver_caller.major == app_version.major &&
-             ver_caller.minor == app_version.minor &&
-             ver_caller.aux >    app_version.aux) ||
-
-            (ver_caller.major == app_version.major &&
-             ver_caller.minor == app_version.minor &&
-             ver_caller.aux ==   app_version.aux &&
-             ver_caller.patch >  app_version.patch)) {
+        if (khm_compare_version(&ver_caller, &app_version) > 0) {
 
             papp_ver->request_swap = TRUE;
 
@@ -195,13 +183,50 @@ khm_ui_cb(LPARAM lParam) {
 #endif
 
     /* make the call */
-    if (!IsBadCodePtr(pcbdata->cb))
-        pcbdata->rv = (*pcbdata->cb)(khm_hwnd_main, pcbdata->rock);
-    else {
-#ifdef DEBUG
-        assert(FALSE);
-#endif
-        pcbdata->rv = KHM_ERROR_INVALID_PARAM;
+    pcbdata->rv = (*pcbdata->cb)(khm_hwnd_main, pcbdata->rock);
+}
+
+
+static void 
+main_wnd_save_sizepos() {
+    RECT r;
+    khm_handle csp_cw;
+    khm_handle csp_mw;
+    const wchar_t * wconfig;
+
+    KillTimer(khm_hwnd_main, MW_RESIZE_TIMER);
+
+    if (khm_main_wnd_mode == KHM_MAIN_WND_MINI)
+        wconfig = L"Windows\\MainMini";
+    else
+        wconfig = L"Windows\\Main";
+
+    GetWindowRect(khm_hwnd_main, &r);
+
+    if (KHM_SUCCEEDED(khc_open_space(NULL,
+                                     L"CredWindow",
+                                     KHM_PERM_WRITE,
+                                     &csp_cw))) {
+        if (KHM_SUCCEEDED(khc_open_space(csp_cw,
+                                        wconfig,
+                                        KHM_PERM_WRITE,
+                                        &csp_mw))) {
+            khm_int32 t;
+
+            khc_write_int32(csp_mw, L"XPos", r.left);
+            khc_write_int32(csp_mw, L"YPos", r.top);
+            khc_write_int32(csp_mw, L"Width", r.right - r.left);
+            khc_write_int32(csp_mw, L"Height", r.bottom - r.top);
+
+            if (KHM_SUCCEEDED(khc_read_int32(csp_mw, L"Dock", &t)) && 
+                t != KHM_DOCK_NONE) {
+                khc_write_int32(csp_mw, L"Dock", KHM_DOCK_AUTO);
+            }
+
+            khc_close_space(csp_mw);
+        }
+
+        khc_close_space(csp_cw);
     }
 }
 
@@ -382,7 +407,13 @@ khm_main_wnd_proc(HWND hwnd,
             return 0;
 
             /* layout control */
+
+        case KHUI_ACTION_VIEW_ALL_IDS:
+            return SendMessage(khm_hwnd_main_cred, uMsg, 
+                               wParam, lParam);
+
         case KHUI_ACTION_LAYOUT_MINI:
+
             if (khm_main_wnd_mode == KHM_MAIN_WND_MINI) {
                 khm_set_main_window_mode(KHM_MAIN_WND_NORMAL);
             } else {
@@ -567,46 +598,7 @@ khm_main_wnd_proc(HWND hwnd,
 
     case WM_TIMER:
         if (wParam == MW_RESIZE_TIMER) {
-            RECT r;
-            khm_handle csp_cw;
-            khm_handle csp_mw;
-            const wchar_t * wconfig;
-
-            if (khm_main_wnd_mode == KHM_MAIN_WND_MINI)
-                wconfig = L"Windows\\MainMini";
-            else
-                wconfig = L"Windows\\Main";
-
-            KillTimer(hwnd, wParam);
-
-            GetWindowRect(hwnd, &r);
-
-            if (KHM_SUCCEEDED(khc_open_space(NULL,
-                                             L"CredWindow",
-                                             KHM_PERM_WRITE,
-                                             &csp_cw))) {
-                if (KHM_SUCCEEDED(khc_open_space(csp_cw,
-                                                 wconfig,
-                                                 KHM_PERM_WRITE,
-                                                 &csp_mw))) {
-                    khm_int32 t;
-
-                    khc_write_int32(csp_mw, L"XPos", r.left);
-                    khc_write_int32(csp_mw, L"YPos", r.top);
-                    khc_write_int32(csp_mw, L"Width",
-                                    r.right - r.left);
-                    khc_write_int32(csp_mw, L"Height",
-                                    r.bottom - r.top);
-
-                    if (KHM_SUCCEEDED(khc_read_int32(csp_mw, L"Dock", &t)) &&
-                        t != KHM_DOCK_NONE) {
-                        khc_write_int32(csp_mw, L"Dock", KHM_DOCK_AUTO);
-                    }
-
-                    khc_close_space(csp_mw);
-                }
-                khc_close_space(csp_cw);
-            }
+            main_wnd_save_sizepos();
 
             return 0;
 
@@ -677,14 +669,13 @@ khm_main_wnd_proc(HWND hwnd,
 
             return kmq_wm_end(m, rv);
         }
-        return 0;
 
     case WM_KHUI_ASSIGN_COMMANDLINE_V1:
         {
             HANDLE hmap;
             void * xfer;
             wchar_t mapname[256];
-            struct tag_khm_startup_options_v1 * pv1opt;
+            khm_startup_options_v1 * pv1opt;
             int code = KHM_ERROR_SUCCESS;
 
             StringCbPrintf(mapname, sizeof(mapname),
@@ -699,7 +690,7 @@ khm_main_wnd_proc(HWND hwnd,
                                  sizeof(*pv1opt));
 
             if (xfer) {
-                pv1opt = (struct tag_khm_startup_options_v1 *) xfer;
+                pv1opt = (khm_startup_options_v1 *) xfer;
 
                 khm_startup.init = pv1opt->init;
                 khm_startup.import = pv1opt->import;
@@ -711,6 +702,7 @@ khm_main_wnd_proc(HWND hwnd,
 
                 khm_startup.no_main_window = FALSE;
                 khm_startup.remote_exit = FALSE;
+                khm_startup.display = 0;
 
                 UnmapViewOfFile(xfer);
             } else {
@@ -743,7 +735,8 @@ khm_main_wnd_proc(HWND hwnd,
             HANDLE hmap;
             void * xfer;
             wchar_t mapname[256];
-            struct tag_khm_startup_options_v2 *pv2opt;
+            khm_startup_options_v2 *pv2opt = NULL;
+            khm_startup_options_v3 *pv3opt = NULL;
             int code = KHM_ERROR_SUCCESS;
 
             StringCbPrintf(mapname, sizeof(mapname),
@@ -758,10 +751,11 @@ khm_main_wnd_proc(HWND hwnd,
                                  sizeof(*pv2opt));
 
             if (xfer) {
-                pv2opt = (struct tag_khm_startup_options_v2 *) xfer;
+                pv2opt = (khm_startup_options_v2 *) xfer;
 
                 if (pv2opt->magic != STARTUP_OPTIONS_MAGIC ||
-                    pv2opt->cb_size != sizeof(*pv2opt)) {
+                    (pv2opt->cb_size != sizeof(*pv2opt) &&
+                     pv2opt->cb_size != sizeof(*pv3opt))) {
                     code = KHM_ERROR_INVALID_PARAM;
                     goto done_with_v2_opt;
                 }
@@ -776,6 +770,14 @@ khm_main_wnd_proc(HWND hwnd,
 
                 pv2opt->code = KHM_ERROR_SUCCESS;
 
+                if (pv2opt->cb_size == sizeof(*pv3opt)) {
+                    pv3opt = (khm_startup_options_v3 *) xfer;
+
+                    khm_startup.display = pv3opt->remote_display;
+                } else {
+                    khm_startup.display = 0;
+                }
+
             done_with_v2_opt:
                 UnmapViewOfFile(xfer);
             } else {
@@ -1114,6 +1116,15 @@ khm_set_main_window_mode(int mode) {
 
     khui_refresh_actions();
 
+    /* 
+     * set the window position before the global khm_main_wnd_mode 
+     * is updated.  otherwise, the windows position for the wrong
+     * mode will be set.  Do not set the window position if the 
+     * main application window has not yet been created.
+     */
+    if (khm_hwnd_main)
+        main_wnd_save_sizepos();
+
     khm_main_wnd_mode = mode;
     if (khm_hwnd_main) {
         khm_get_main_window_rect(&r);
@@ -1201,18 +1212,29 @@ khm_show_main_window(void) {
         if (hw != khm_hwnd_main)
             SetForegroundWindow(khm_hwnd_main);
     }
-
-    if (khm_nCmdShow == SW_SHOWMINIMIZED ||
+    /*
+     * We test for the values of khm_nCmdShow that
+     * can be set at process startup.  They will 
+     * only be seen the first time this function is 
+     * called.  After the first time, the value of
+     * khm_nCmdShow will always be SW_RESTORE.
+     * When one of the minimized values is set, 
+     * khm_show_main_window() will not be called
+     * unless the user initiates a request to show
+     * the window.
+     */
+    else if (khm_nCmdShow == SW_SHOWMINIMIZED ||
         khm_nCmdShow == SW_SHOWMINNOACTIVE ||
         khm_nCmdShow == SW_MINIMIZE) {
-        khm_hide_main_window();
-    } else {
-        ShowWindow(khm_hwnd_main, khm_nCmdShow);
-        UpdateWindow(khm_hwnd_main);
 
-        khm_cred_refresh();
+        khm_nCmdShow = SW_SHOW;
     }
 
+    ShowWindow(khm_hwnd_main, khm_nCmdShow);
+    UpdateWindow(khm_hwnd_main);
+    
+    khm_cred_refresh();
+
     khm_nCmdShow = SW_RESTORE;
 }
 
@@ -1327,7 +1349,7 @@ khm_register_main_wnd_class(void) {
     wc.cbWndExtra = 0;
     wc.hInstance = khm_hInstance;
     wc.hIcon = LoadIcon(khm_hInstance, MAKEINTRESOURCE(IDI_MAIN_APP));
-    wc.hCursor = LoadCursor((HINSTANCE) NULL, MAKEINTRESOURCE(IDC_ARROW));
+    wc.hCursor = LoadCursor((HINSTANCE) NULL, IDC_ARROW);
     wc.hIconSm = LoadImage(khm_hInstance, MAKEINTRESOURCE(IDI_MAIN_APP), IMAGE_ICON, 0, 0, LR_DEFAULTSIZE);
     wc.hbrBackground = (HBRUSH) (COLOR_APPWORKSPACE);
     wc.lpszMenuName = NULL;
@@ -1342,7 +1364,7 @@ khm_register_main_wnd_class(void) {
     wc.cbWndExtra = 0;
     wc.hInstance = khm_hInstance;
     wc.hIcon = LoadIcon(khm_hInstance, MAKEINTRESOURCE(IDI_MAIN_APP));
-    wc.hCursor = LoadCursor((HINSTANCE) NULL, MAKEINTRESOURCE(IDC_ARROW));
+    wc.hCursor = LoadCursor((HINSTANCE) NULL, IDC_ARROW);
     wc.hIconSm = LoadImage(khm_hInstance, MAKEINTRESOURCE(IDI_MAIN_APP), IMAGE_ICON, 0, 0, LR_DEFAULTSIZE);
     wc.hbrBackground = (HBRUSH) (COLOR_APPWORKSPACE);
     wc.lpszMenuName = NULL;
